@@ -20,28 +20,31 @@
 
   Minimization code follows the same overall structure as optimizers in JAX.
   Optimizers return two functions:
-    init_fn: 
+    init_fn:
       Function that initializes the  state of an optimizer. Should take
       positions as an ndarray of shape `[n, output_dimension]`. Returns a state
       which will be a namedtuple.
-    apply_fn: 
+    apply_fn:
       Function that takes a state and produces a new state after one
       step of optimization.
 """
 
 from collections import namedtuple
 
-from typing import TypeVar, Callable, Tuple, Union
+from typing import TypeVar, Callable, Tuple, Union, Any
 
 import jax.numpy as jnp
+from jax.tree_util import tree_map, tree_reduce
 
 from jax_md import quantity
 from jax_md import dataclasses
 from jax_md import util
 from jax_md import space
+from jax_md import simulate
 
 # Types
 
+PyTree = Any
 Array = util.Array
 f32 = util.f32
 f64 = util.f64
@@ -99,8 +102,9 @@ class FireDescentState:
     n_pos: The number of steps in the right direction, so far.
   """
   position: Array
-  velocity: Array
+  momentum: Array
   force: Array
+  mass: Array
   dt: float
   alpha: float
   n_pos: int
@@ -117,14 +121,15 @@ def fire_descent(energy_or_force: Callable[..., Array],
                  f_alpha: float=0.99) -> Minimizer[FireDescentState]:
   """Defines FIRE minimization.
 
-  This code implements the "Fast Inertial Relaxation Engine" from Bitzek et al. [#bitzek]_
+  This code implements the "Fast Inertial Relaxation Engine" from Bitzek et
+  al. [#bitzek]_
 
   Args:
     energy_or_force: A function that produces either an energy or a force from
       a set of particle positions specified as an ndarray of shape
       `[n, spatial_dimension]`.
-    shift_fn: A function that displaces positions `R`, by an amount `dR`. Both `R`
-      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
+    shift_fn: A function that displaces positions `R`, by an amount `dR`. Both
+      `R` and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
     quant: Either a quantity.Energy or a quantity. Force specifying whether
       energy_or_force is an energy or force respectively.
     dt_start: The initial step size during minimization as a float.
@@ -145,50 +150,57 @@ def fire_descent(energy_or_force: Callable[..., Array],
       and Peter Gumbsch. "Structural relaxation made simple."
       Physical review letters 97, no. 17 (2006): 170201.
   """
-
   dt_start, dt_max, n_min, f_inc, f_dec, alpha_start, f_alpha = util.static_cast(
     dt_start, dt_max, n_min, f_inc, f_dec, alpha_start, f_alpha)
 
+  nve_init_fn, nve_step_fn = simulate.nve(energy_or_force, shift_fn, dt_start)
   force = quantity.canonicalize_force(energy_or_force)
-  def init_fn(R: Array, **kwargs) -> FireDescentState:
-    V = jnp.zeros_like(R)
+
+  def init_fn(R: PyTree, mass: Array=1.0, **kwargs) -> FireDescentState:
+    P = tree_map(lambda x: jnp.zeros_like(x), R)
     n_pos = jnp.zeros((), jnp.int32)
     F = force(R, **kwargs)
-    return FireDescentState(R, V, F, dt_start, alpha_start, n_pos)  # pytype: disable=wrong-arg-count
+    mass = simulate.canonicalize_mass(mass)
+    return FireDescentState(R, P, F, mass, dt_start, alpha_start, n_pos)  # pytype: disable=wrong-arg-count
+
   def apply_fn(state: FireDescentState, **kwargs) -> FireDescentState:
-    R, V, F_old, dt, alpha, n_pos = dataclasses.astuple(state)
-
-    R = shift_fn(R, dt * V + dt ** f32(2) * F_old, **kwargs)
-
-    F = force(R, **kwargs)
-
-    V = V + dt * f32(0.5) * (F_old + F)
+    state = nve_step_fn(state, dt=state.dt, **kwargs)
+    R, P, F, M, dt, alpha, n_pos = dataclasses.unpack(state)
 
     # NOTE(schsam): This will be wrong if F_norm ~< 1e-8.
     # TODO(schsam): We should check for forces below 1e-6. @ErrorChecking
-    F_norm = jnp.sqrt(jnp.sum(F ** f32(2)) + f32(1e-6))
-    V_norm = jnp.sqrt(jnp.sum(V ** f32(2)))
+    F_norm = jnp.sqrt(tree_reduce(lambda accum, f:
+                                  accum + jnp.sum(f ** 2) + 1e-6, F, 0.0))
+    P_norm = jnp.sqrt(tree_reduce(lambda accum, p:
+                                  accum + jnp.sum(p ** 2), P, 0.0))
 
-    P = jnp.array(jnp.dot(jnp.reshape(F, (-1)), jnp.reshape(V, (-1))))
-
-    V = V + alpha * (F * V_norm / F_norm - V)
+    # NOTE: In the original FIRE algorithm, the quantity that determines when
+    # to reset the momenta is F.V rather than F.P. However, all of the JAX MD
+    # simulations are in momentum space for easier agreement with prior work /
+    # rigid body physics. We only use the sign of F.P here, which shouldn't
+    # differ from F.V, however if there are regressions then we should
+    # reconsider this choice.
+    F_dot_P = tree_reduce(
+        lambda accum, f_dot_p: accum + f_dot_p,
+        tree_map(lambda f, p: jnp.sum(f * p), F, P))
+    P = tree_map(lambda p, f: p + alpha * (f * P_norm / F_norm - p), P, F)
 
     # NOTE(schsam): Can we clean this up at all?
-    n_pos = jnp.where(P >= 0, n_pos + 1, 0)
+    n_pos = jnp.where(F_dot_P >= 0, n_pos + 1, 0)
     dt_choice = jnp.array([dt * f_inc, dt_max])
-    dt = jnp.where(P > 0,
+    dt = jnp.where(F_dot_P > 0,
                    jnp.where(n_pos > n_min,
                              jnp.min(dt_choice),
                              dt),
                    dt)
-    dt = jnp.where(P < 0, dt * f_dec, dt)
-    alpha = jnp.where(P > 0,
+    dt = jnp.where(F_dot_P < 0, dt * f_dec, dt)
+    alpha = jnp.where(F_dot_P > 0,
                       jnp.where(n_pos > n_min,
                                 alpha * f_alpha,
                                 alpha),
                       alpha)
-    alpha = jnp.where(P < 0, alpha_start, alpha)
-    V = (P < 0) * jnp.zeros_like(V) + (P >= 0) * V
+    alpha = jnp.where(F_dot_P < 0, alpha_start, alpha)
+    P = tree_map(lambda p: (F_dot_P >= 0) * p, P)
 
-    return FireDescentState(R, V, F, dt, alpha, n_pos)  # pytype: disable=wrong-arg-count
+    return FireDescentState(R, P, F, M, dt, alpha, n_pos)  # pytype: disable=wrong-arg-count
   return init_fn, apply_fn
